/*
 *  $Id: cfg.c,v 1.6 1999/05/12 08:27:40 lord Exp $
 *
 *  Vadim Zaliva <lord@crocodile.org>
 *  http://www.crocodile.org/
 *
 *  Copyright (C) 1999 Vadim Zaliva
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "cfg.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#define MAXCFGLINE 1024

static int  cfg_entry_cmp(const void *a,const void *b);

char *cfgfind(char *name,struct Cfg *cfg, int offset)
{
    int i,j;
    struct Dict *res;
    struct Dict sample;

    sample.name=name;
    sample.value=NULL;

    res=(struct Dict *)bsearch(&sample,
                               cfg->dict,
                               0,
                               sizeof(struct Dict *),
                               cfg_entry_cmp
    );
    
    if(!res)
        return NULL;
    
    for(j=0;j<=offset;j++)
    {
        if(cfg->dict[i]->value[j]==NULL)
            return NULL;
    }
    return cfg->dict[i]->value[offset];
}

int writecfg(const char *name,struct Cfg *cfg)
{
    FILE *f;
    int  j;
    int i;

    if((f=fopen(name,"wb"))==NULL)
        return -1;

    if(cfg)
    {
        for(i=0;i<cfg->nelements;i++)
        {
            if(!cfg->dict[i]->name) return;
            fprintf(f,"%s\t",cfg->dict[i]->name);
            for(j=0;j<cfg->dict[i]->nvalues;j++)
                if(cfg->dict[i]->value[j])
                    fprintf(f,"\t%s",cfg->dict[i]->value[j]);
            
            fprintf(f,"\n");
        }
    }
    fclose(f);
    return 0;
}

void freecfg(struct Cfg *cfg)
{
    int i,j;

    if(!cfg)
        return;
    
    for(i=0;i<cfg->nelements;i++)
    {
        if(cfg->dict[i]->value)
        {
            for(j=0;j<cfg->dict[i]->nvalues;j++)
                if(cfg->dict[i]->value[j])
                    free(cfg->dict[i]->value[j]);
            free(cfg->dict[i]->value);
        }
        
        if(cfg->dict[i]->name)
            free(cfg->dict[i]->name);
    }
    
    free(cfg->dict);
    
    free(cfg);
}

struct Cfg *readcfg(const char *name)
{
    struct Cfg  *cfg  ;
    FILE       *f;
    char        tmp[MAXCFGLINE];
    char       *s;
    int         c;
    
    enum 
    {
        START,
        NAME,
        VALUE,
        INQUOTE,
        WHITESPACE,
        COMMENT
    } state=START;
    
    int n,i;
    char *pname,*pvalue;
    
    if((f=fopen(name,"rb"))==NULL)
        return NULL;

    cfg=malloc(sizeof(struct Cfg));
    cfg->nelements=0;
    s=tmp;
    
    while((c=fgetc(f))!=EOF)
    {
        /* Order of 'case' statements is important here! */
        switch(state)
        {
        case START:
            if(c=='#')
            {
                state=COMMENT;
                break;
            }
            else
                if(!isspace(c))
                {
                    s=tmp;
                    state=NAME;
                } else
                {
                    break;
                }
            
        case NAME:
            if(isspace(c))
            {
                struct Dict *tmp1=malloc(sizeof(struct Dict));
                *s='\0';
                tmp1->nvalues = 0;
                tmp1->value   = NULL;
                tmp1->name    = strdup(tmp);
                if(cfg->nelements)
                {
                    struct Dict **last=cfg->dict;
                    cfg->dict=malloc(sizeof(struct Dict *)*(cfg->nelements+1));
                    memcpy(cfg->dict,last,sizeof(struct Dict *)*cfg->nelements);
                    cfg->dict[cfg->nelements]=tmp1;
                    cfg->nelements++;
                }
                else
                {
                    cfg->dict=malloc(sizeof(struct Dict *));
                    cfg->dict[0]=tmp1;
                    cfg->nelements=1;
                }
                state=WHITESPACE;
            }
            else
            {
                *s++=c;
            }
            break;

        case WHITESPACE:
            if(c=='\n')
            {
                state=START;
                break;
            }
            else
            {
                if(!isspace(c))
                {
                    s=tmp;
                    state=VALUE;
                } else
                {
                    break;
                }
            }
            
        case VALUE:
            if(c=='"')
                state=INQUOTE;
            else
                if(isspace(c))
                {
                    struct Dict *last=cfg->dict[cfg->nelements-1];
                    char **tmp1;
                    int  i;
                    
                    *s='\0';
                    tmp1=last->value;
                    last->value=malloc((last->nvalues+1)*sizeof(char *));
                    if(tmp1)
                    {
                        for(i=0;i<last->nvalues;i++)
                            last->value[i]=tmp1[i];
                        free(tmp1);
                    }
                    last->value[last->nvalues]=strdup(tmp);
                    last->nvalues++;
                    if(c=='\n')
                        state=START;
                    else
                        state=WHITESPACE;
                } else
                {
                    *s++=c;
                }
            break;

        case INQUOTE:
            if(c=='"')
                state=VALUE;
            else
                *s++=c;
            break;

        case COMMENT:
            if(c=='\n')
                state=START;
            break;
        }
    }

    qsort((void *) cfg->dict,
          cfg->nelements,
          sizeof(struct Dict *),
          cfg_entry_cmp);

    return cfg;
}

static int cfg_entry_cmp(const void *a,const void *b)
{
    return(strcmp(
	((struct Dict *) a)->name,
        ((struct Dict *) b)->name
    ));
}


